/*
 Illinois Open Source License

 University of Illinois/NCSA
 Open Source License

 Copyright © 2009,    University of Illinois.  All rights reserved.

 Developed by:
 Innovative Systems Lab
 National Center for Supercomputing Applications
 http://www.ncsa.uiuc.edu/AboutUs/Directorates/ISL.html

 Permission is hereby granted, free of charge, to any person obtaining a copy of this 
 software and associated documentation files (the "Software"), to deal with the Software 
 without restriction, including without limitation the rights to use, copy, modify, merge, 
 publish, distribute, sublicense, and/or sell copies of the Software, and to permit 
 persons to whom the Software is furnished to do so, subject to the following conditions:
 
 * Redistributions of source code must retain the above copyright notice, this list of 
   conditions and the following disclaimers.
 
 * Redistributions in binary form must reproduce the above copyright notice, this list of 
   conditions and the following disclaimers in the documentation and/or other materials 
   provided with the distribution.
 
 * Neither the names of Innovative Systems Lab and National Center for Supercomputing 
   Applications, nor the names of its contributors may be used to endorse or promote 
   products derived from this Software without specific prior written permission.
 
   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
   INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
   PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE 
   LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT 
   OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
   DEALINGS WITH THE SOFTWARE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <math.h>
#include <strings.h>
#include <malloc.h>

#include "args.h"
#include "model.h"
#include <libspe2.h>
#include <pputask.h>
#include "spu/compute_task.h"


void* memset(void*, int, size_t);

#define MAXSPES 16
extern spe_program_handle_t spu_main;

long long DD[MAXNJK][MAXBINS] __attribute__ ((aligned (128)));
long long RRS_BUF[MAXSPES][MAXNJK][MAXBINS] __attribute__ ((aligned (128)));
long long DRS_BUF[MAXSPES][MAXNJK][MAXBINS] __attribute__ ((aligned (128)));
long long RRS[MAXNJK][MAXBINS] __attribute__ ((aligned (128)));
long long DRS[MAXNJK][MAXBINS] __attribute__ ((aligned (128)));

#define TDIFF(ts, te)  (te.tv_sec - ts.tv_sec + (te.tv_usec - ts.tv_usec) * 1e-6)

double *init_bins(int bins_per_dec, float min_angle, float max_angle, int angle_units, int *nbins);

extern int upload_binbs(double* binb, int nbins);

void
displayDD()
{
    int i, j;
    for (i = 0;i < 11; i++){
	for (j =0 ; j < MAXBINS; j++){
	    printf("%lld  ", DD[i][j]);	    
	}
	printf("\n");
    }
    
    exit(1);
}


void write_results(long long DD[][MAXBINS], long long RRS[][MAXBINS], long long DRS[][MAXBINS], 
		   int njk, int nbins, int bins_per_dec, float min_angle, int random_count, char *fname)
{
    // compute and output results
  FILE *outfile;
  int k, l;
  
  if ((outfile = fopen(fname, "w")) == NULL)
    {
      fprintf(stderr, "Unable to open output file %s for writing, assuming stdout\n", fname);
      outfile = stdout;
    }
  
  fprintf(outfile, "boundary DD DRS RRS w ");
  
  for (l = 1; l < njk+1; l++)
    fprintf(outfile, "DD[%i] DRS[%i] RRS w[%i]  ", l, l, l);
  
  fprintf(outfile, "\n");
  
  double bb = pow(10, log10(min_angle) + 0*1.0/bins_per_dec);
  fprintf(outfile, "<%f %lld %lld %lld -\n", bb, DD[0][0], DRS[0][0], RRS[0][0]);
  
  for (k = 1; k < nbins+1; k++)
    {
      double bb = pow(10, log10(min_angle) + k*1.0/bins_per_dec);
      fprintf(outfile, "%f ", bb);


      for (l = 0; l < njk+1; l++)
        {
          if (RRS[0][k] != 0)
	    {
	      double norm_dd = (double)DD[l][k]; // (args.ndpoints * (args.ndpoints - 1) / 2.0);
	      double norm_rrs = (double)RRS[0][k]; // (args.random_count * args.nrpoints * (args.nrpoints - 1) / 2.0);
	      double norm_drs = (double)DRS[l][k]; // (args.random_count * args.ndpoints * args.nrpoints);
	      double w = (random_count * norm_dd - norm_drs) / norm_rrs + 1.0;
	      //printf("%f\t", w);
	      fprintf(outfile, "%lld %lld %lld %f ", DD[l][k], DRS[l][k], RRS[0][k], w);
	    }
          else{
	      //printf("%lld %lld %lld - ", DD[l][k], DRS[l][k], RRS[0][k]);
	      fprintf(outfile, "%lld %lld %lld - ", DD[l][k], DRS[l][k], RRS[0][k]);
	  }
        }
      
      fprintf(outfile, "\n");
      //printf("\n");
    }
  
  bb = pow(10, log10(min_angle) + nbins*1.0/bins_per_dec);
  fprintf(outfile, ">%f %lld %lld %lld\n", bb, DD[0][nbins+1], DRS[0][nbins+1], RRS[0][nbins+1]);
  
  fclose(outfile);
}


void verify_results(long long DD[][MAXBINS], long long RRS[][MAXBINS], long long DRS[][MAXBINS],
                   int njk, int nbins, int bins_per_dec, float min_angle, int random_count);

int main(int argc, char *argv[])
{
    int rf;
    int numspes = 8;
    int i, j, k;

      
    memset(RRS_BUF, 0, sizeof(RRS_BUF));
    memset(DRS_BUF, 0, sizeof(DRS_BUF));
    memset(RRS, 0, sizeof(RRS));
    memset(DRS, 0, sizeof(DRS));
    memset(DD, 0, sizeof(DD));
    
	//parse command line args
	options args;
	parse_args(argc, argv, &args);
	
	ppu_task_spe_num_set(numspes);
	ppu_task_init(argc, argv, spu_main);
	
	printf("\ndata file: %s \n", args.data_name);
	printf("number of data points: %d\n", args.ndpoints);
	//printf("data points per segment: %d\n", args.maxdsize);
	
	printf("random data files: %s.1-%i\n", args.random_name, args.random_count);
	printf("number of random points: %d \n", args.nrpoints);
	//printf("random points per segment: %d\n", args.maxrsize);
	//printf("compute RRS: %s\n", (args.doRRS) ? "YES" : "NO");
	
	printf("output file: %s \n", args.output_name);

	int nbins;
	double *binb = init_bins(args.bins_per_dec, args.min_angle, args.max_angle, args.angle_units, &nbins);
	

	// memory for input data
	struct cartesian *data = (struct cartesian*)memalign(128,ROUNDUP_32(args.ndpoints)*sizeof(struct cartesian));
	struct cartesian2 *data_v = (struct cartesian2*)memalign(128,ROUNDUP_32(args.ndpoints)*sizeof(struct cartesian2)/2);
	if (data == NULL || data_v == NULL)
	    {
		fprintf(stderr, "Unable to allocate memory for % data points (#1)\n", args.ndpoints);
		exit(0);
	  }
	
	struct cartesian* random[numspes];
	struct cartesian2* random_v[numspes];
	
	for(i=0; i < numspes; i++){
	    random[i] = (struct cartesian*)memalign(128, ROUNDUP_32(args.nrpoints)*sizeof(struct cartesian));
	    random_v[i] = (struct cartesian2*)memalign(128, ROUNDUP_32(args.nrpoints)*sizeof(struct cartesian2)/2);       
	    if (random[i] == NULL || random_v[i] == NULL)
		{
		    fprintf(stderr, "Unable to allocate memory for % data points (#2)\n", args.nrpoints);
		    exit(0);
		}
	}
	
	struct timeval t2, t1, t0;
	float t_computeDD=0, t_computeRRS=0, t_computeDRS=0, t_fileIO=0;
	
	int npd=0;
	int npr=0;
	
	// loop over all files
	npd = readdatafile(args.data_name, data, data_v, args.ndpoints);
	if (npd != args.ndpoints)
	    {
		fprintf(stderr, "Error: read %i data points out of %i\n", npd, args.ndpoints);
		exit(0);
	    }
	
	
	assign_jk(data, npd, args.njk);
	t_computeDD = 0;
	
	upload_binbs(binb, nbins);


	if (numspes == 1){
	    gettimeofday(&t0, NULL);
	    doComputeWaterfall(data, npd, data, data_v, npd, 1, DD, args.njk+1, FLAG_DD, 0);
	    //doComputeWaterfall_cpu(data, npd, data, npd, 1, DD, nbins, binb,args.njk+1);
	    ppu_task_spu_waitall();
	    gettimeofday(&t1, NULL);
	    t_computeDD += TDIFF(t0, t1);
	}
	

	for (rf = 1; rf < args.random_count+1; rf+=numspes){

	    int active_spes = numspes;
	    if ( (rf + numspes) > (args.random_count +1)){
		active_spes = (rf + numspes) - (args.random_count +1);
	    }
	    if (numspes == 1){
		if (rf % 10 == 0){
		    printf("Processing dataset # %i...\n", rf);
		}
	    }else{
		printf("Processing dataset # %i...\n", rf);
	    }
	    fflush(stdout);
	    for (i=0;i < active_spes; i++){
		gettimeofday(&t0, NULL);	    	    
		char fname[256];	    
		sprintf(fname, "%s.%i", args.random_name, rf+i);
		
		npr = readdatafile(fname, random[i], random_v[i], args.nrpoints);
		if (npr != args.nrpoints){
		    fprintf(stderr, "Error: read %i random points out of %i in file %s\n", npr, args.nrpoints, fname);
		    exit(0);
		}
	    
		gettimeofday(&t1, NULL);
		t_fileIO += TDIFF(t0, t1);
	    }
	    
	    gettimeofday(&t0, NULL);
	    
	    if(active_spes < numspes){
		doComputeWaterfall(data, npd, data, data_v, npd, 1, DD, args.njk+1, FLAG_DD, active_spes);
		//doComputeWaterfall_cpu(data, npd, data, npd, 1, DD, nbins, binb,args.njk+1);
	    }
	    for (i=0;i < active_spes; i++){
		doComputeWaterfall(random[i], npr, random[i], random_v[i], npr, 1, RRS_BUF[i], 1, FLAG_RR, i);
		//doComputeWaterfall_cpu(random[i], npr, random[i],  npr, 1, RRS_BUF[i], nbins, binb, 1);
	    }
	    
	    ppu_task_spu_waitall();
	    
	    gettimeofday(&t1, NULL);
	    
	    for (i=0;i < active_spes; i++){
		doComputeWaterfall(data, npd, random[i], random_v[i],  npr, 0, DRS_BUF[i], args.njk+1, FLAG_DR, i);
		//doComputeWaterfall_cpu(data, npd, random[i], npr, 0, DRS_BUF[i], nbins, binb, args.njk+1);
	    }
	    ppu_task_spu_waitall();
	    
	    gettimeofday(&t2, NULL);
	    
	    t_computeRRS += TDIFF(t0, t1);
	    t_computeDRS += TDIFF(t1, t2);
	    
	}
	

	for (i = 0;i < numspes; i++){
	    for (j =0; j < MAXNJK; j++){
		for (k =0;k < MAXBINS;k++){
		    RRS[j][k] += RRS_BUF[i][j][k];
		    DRS[j][k] += DRS_BUF[i][j][k];
		}
	    }
	}
	

	write_results(DD, RRS, DRS, args.njk, nbins, args.bins_per_dec, args.min_angle, args.random_count, args.output_name);
	
	double mytime = t_computeDD+t_computeRRS+t_computeDRS;
	printf("================================================\n");
	printf("Time to compute DD: %.4f sec\n", t_computeDD);
	printf("Time to compute RRS: %.4f sec\n", t_computeRRS);
	printf("Time to compute DRS: %.4f sec\n", t_computeDRS);
	printf("Time to load data files: %.4f sec\n", t_fileIO);
	printf("Time to compute DD, RRS, & DRS: %.4f sec\n", t_computeDD+t_computeRRS+t_computeDRS);
	printf("TOTAL time (DD+RRS+DRS+IO): %.4f sec\n", t_computeDD+t_computeRRS+t_computeDRS+t_fileIO);
	printf("================================================\n");
	printf("Mytime is %.4f sec (p=%d, r=%d)\n", mytime, args.ndpoints, args.nrpoints);
	
	//verify_results(DD, RRS, DRS, args.njk, nbins, args.bins_per_dec, args.min_angle, args.random_count);
	
	// free memory
	free((char *)data);
	for (i =0; i< numspes; i++){
	    free(random[i]);
	    free(random_v[i]);
	}
	free((char *)binb);
	

	//printf("DRS compute time =%.4f sec\n", t_computeDRS); 
	return 0;
}




double *init_bins(int bins_per_dec, float min_angle, float max_angle, int angle_units, int *nbins)
{
    int k;
    
    *nbins = floor(bins_per_dec * (log10(max_angle) - log10(min_angle)));
    
    // memory for bin boundaries
    printf("nbins=%d\n", *nbins);
    int size = ROUNDUP_32( ((*nbins)+1) * sizeof(double));
    printf("allocating %d bytes for bingbo in host\n", size);
    double *binb = (double *)memalign(128, size);
    if (binb == NULL)
      {
        fprintf(stderr, "Unable to allocate memory for bin boundaries\n");
        exit(0);
      }

    printf("binb=%p\n", binb);
    printf("Min angualr distance: %f %s\n", min_angle, (angle_units) ? "arcmin" : "degrees");
    printf("Max angular distance: %f %s\n", max_angle, (angle_units) ? "arcmin" : "degrees");
    printf("Bins per dec: %i\n", bins_per_dec);
    printf("Total bins  : %i\n", *nbins);

    printf("Bin boundaries:\n");
    printf("#  degrees          arcminutes        dot_product\n");
    for (k = 0; k < (*nbins)+1; k++)
    {
        double bb  = pow(10, log10(min_angle) + k*1.0/bins_per_dec);
        binb[k] = cos(bb / ((angle_units) ? 60.0 : 1.0) * d2r);

        printf("%i %.15f %.15f ", k, (angle_units) ? bb/60.0f : bb, (angle_units) ? bb : bb * 60);
        printf("%.15f\n", binb[k]);
    }
    printf("\n");

    return binb;
}

